package org.apache.ojb.broker.accesslayer;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.apache.ojb.broker.platforms.Platform;
import org.apache.ojb.broker.platforms.PlatformException;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.logging.LoggerFactory;

import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.SQLException;

/**
 * Intern used wrapper for {@link Statement} and {@link ResultSet} instances.
 *
 * @version $Id: ResultSetAndStatement.java,v 1.13 2004/05/22 09:51:25 brj Exp $
 */
public class ResultSetAndStatement
{
	private static Logger log = LoggerFactory.getLogger(ResultSetAndStatement.class);

	private final Platform m_platform;
    private boolean isClosed;
    /*
    arminw: declare final to avoid stmt/rs leaking in use
    by re-setting these fields.
    */
	public final ResultSet m_rs;
	public final Statement m_stmt;

	public ResultSetAndStatement(Platform platform, Statement stmt, ResultSet rs)
	{
		m_platform = platform;
        m_stmt = stmt;
        m_rs = rs;
        isClosed = false;
	}

	/**
	 * do a platform specific resource release.
     * <br/>
     * Note: This method must be called after usage
     * of this class.
	 */
	public void close()
	{
		if(isClosed) return;

        try
		{
			// this will close the result set
            m_platform.beforeStatementClose(m_stmt, m_rs);
		}
		catch (PlatformException e)
		{
			log.error("Platform dependent operation failed", e);
		}
        // we want to close the statement in any case
        finally
        {
            try
            {
                isClosed = true;
                //close statement on wrapped statement class, or real statement
                if (m_stmt != null)
                {
                    m_stmt.close();
                }
                m_platform.afterStatementClose(m_stmt, m_rs);
            }
            catch (SQLException ignored)
            {
                if (log.isDebugEnabled()) log.debug("Statement closing failed", ignored);
            }
            catch (PlatformException e)
            {
                log.error("Platform dependent operation failed", e);
            }
        }
	}

    protected void finalize() throws Throwable
    {
        super.finalize();
        if(!isClosed && (m_stmt != null || m_rs != null))
        {
            log.warn("** Associated resources (Statement/ResultSet) not closed!" +
                    " Try automatic cleanup **");
            try
            {
                close();
            }
            catch (Exception ignore)
            {
                //ignore it                
            }
        }
    }
}
