package org.apache.ojb.broker.util.logging;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;

import org.apache.log4j.LogManager;
import org.apache.log4j.Level;
import org.apache.log4j.PropertyConfigurator;
import org.apache.ojb.broker.util.ClassHelper;
import org.apache.ojb.broker.util.configuration.Configuration;
import org.apache.ojb.broker.util.configuration.ConfigurationException;

/**
 * This is a Logger implementation based on Log4j.
 * It can be enabled by putting
 * LoggerClass=org.apache.ojb.broker.util.logging.Log4jLoggerImpl
 * in the OJB .properties file. <br>
 * If you want log4j to initialize from a property file you can add
 * LoggerConfigFile=log4j.properties to the org.apache.ojb.properties file.
 * the logger only initializes log4j if the application hasn't done it yet
 *
 * You can find sample log4j.properties file in the log4j web site
 * http://jakarta.apache.org/log4j
 * in the javadoc look for org.apache.log4j.examples
 *
 * @author Bertrand
 * @author Thomas Mahler
 * @version $Id: Log4jLoggerImpl.java,v 1.16.2.1 2004/11/03 19:11:48 brj Exp $
 */
public class Log4jLoggerImpl implements Logger
{
	static private HashMap priorityMap;
	private transient org.apache.log4j.Logger logger;
	private String name;

	static private final String FQCN = Log4jLoggerImpl.class.getName();

    /** flag about log4j configuration state */
    private static boolean log4jConfigured = false;

	static {
		priorityMap = new HashMap();
		priorityMap.put(new Integer(Logger.DEBUG), Level.DEBUG);
		priorityMap.put(new Integer(Logger.INFO), Level.INFO);
		priorityMap.put(new Integer(Logger.WARN), Level.WARN);
		priorityMap.put(new Integer(Logger.ERROR), Level.ERROR);
		priorityMap.put(new Integer(Logger.FATAL), Level.FATAL);
	}

    /**
     * Helper method to check if log4j is already configured
     */
    private static synchronized boolean isLog4JConfigured()
    {
        if (log4jConfigured)
            return true;

        Enumeration en = org.apache.log4j.Logger.getRootLogger().getAllAppenders();

        if (!(en instanceof org.apache.log4j.helpers.NullEnumeration))
        {
            log4jConfigured = true;
            return true;
        }
        else
        {
            Enumeration cats = LogManager.getCurrentLoggers();
            while (cats.hasMoreElements())
            {
                org.apache.log4j.Logger c = (org.apache.log4j.Logger) cats.nextElement();
                if (!(c.getAllAppenders() instanceof org.apache.log4j.helpers.NullEnumeration))
                {
                    log4jConfigured = true;
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Initialization of log4j <br>
     * <b>NOTE</b>  - if log4j property file is called log4j.properties then
     * log4j will be configured already.
     *
     */
    private static synchronized void initializeLog4JSubSystem(String configFile)
    {
        LoggerFactory.getBootLogger().info("Initializing Log4J using file:" + configFile);
        if (configFile != null)
        {
            // try resource look in classpath
            URL url = ClassHelper.getResource(configFile);
            LoggerFactory.getBootLogger().info("Initializing Log4J : resource from config file:" + url);
            if (url != null)
            {
                PropertyConfigurator.configure(url);
            }

            // if file is not in classpath try ordinary filesystem lookup
            else if (configFile != "")
            {
                PropertyConfigurator.configure(configFile);
            }
            else
            {
                // no configuration available
            }
        }
        else
        {
            // no configuration available
        }
    }

	public Log4jLoggerImpl(String name)
	{
		this.name = name;
		logger = org.apache.log4j.Logger.getLogger(name);
	}

	/**
	 * Gets the logger.
	 * @return Returns a Category
	 */
	private org.apache.log4j.Logger getLogger()
	{
		if (logger == null)
		{
			logger = org.apache.log4j.Logger.getLogger(name);
		}
		return logger;
	}

	public String getName()
	{
		return name;
	}

	private Level getLevel()
	{
		return getLogger().getEffectiveLevel();
	}

	/**
	 * generate a message for loglevel DEBUG
	 * @param pObject the message Object
	 */
	public final void debug(Object pObject)
	{
		getLogger().log(FQCN, Level.DEBUG, pObject, null);
	}

	/**
	 * generate a message for loglevel INFO
	 * @param pObject the message Object
	 */
	public final void info(Object pObject)
	{
		getLogger().log(FQCN, Level.INFO, pObject, null);
	}

	/**
	 * generate a message for loglevel WARN
	 * @param pObject the message Object
	 */
	public final void warn(Object pObject)
	{
		getLogger().log(FQCN, Level.WARN, pObject, null);
	}

	/**
	 * generate a message for loglevel ERROR
	 * @param pObject the message Object
	 */
	public final void error(Object pObject)
	{
		getLogger().log(FQCN, Level.ERROR, pObject, null);
	}

	/**
	 * generate a message for loglevel FATAL
	 * @param pObject the message Object
	 */
	public final void fatal(Object pObject)
	{
		getLogger().log(FQCN, Level.FATAL, pObject, null);
	}

	public void debug(Object message, Throwable obj)
	{
		getLogger().log(FQCN, Level.DEBUG, message, obj);
	}

	public void error(Object message, Throwable obj)
	{
		getLogger().log(FQCN, Level.ERROR, message, obj);
	}

	public void fatal(Object message, Throwable obj)
	{
		getLogger().log(FQCN, Level.FATAL, message, obj);
	}

	public void info(Object message, Throwable obj)
	{
		getLogger().log(FQCN, Level.INFO, message, obj);
	}

	public void warn(Object message, Throwable obj)
	{
		getLogger().log(FQCN, Level.WARN, message, obj);
	}

	public void safeDebug(String message, Object obj)
	{
		if (Level.DEBUG.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			if (obj != null)
			{
				try
				{
					toString = obj.toString();
				}
				catch (Throwable t)
				{
					toString = "BAD toString() impl for " + obj.getClass().getName();
				}
			}
			debug(message + " : " + toString);
		}
	}

	public void safeDebug(String message, Object obj, Throwable throwable)
	{
		if (Level.DEBUG.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			if (obj != null)
			{
				try
				{
					toString = obj.toString();
				}
				catch (Throwable t)
				{
					toString = "BAD toString() impl for " + obj.getClass().getName();
				}
			}
			debug(message + " : " + toString, throwable);
		}
	}

	public void safeInfo(String message, Object obj)
	{
		if (Level.INFO.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			if (obj != null)
			{
				try
				{
					toString = obj.toString();
				}
				catch (Throwable t)
				{
					toString = "BAD toString() impl for " + obj.getClass().getName();
				}
			}
			info(message + " : " + toString);
		}
	}

	public void safeInfo(String message, Object obj, Throwable throwable)
	{
		if (Level.INFO.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			if (obj != null)
			{
				try
				{
					toString = obj.toString();
				}
				catch (Throwable t)
				{
					toString = "BAD toString() impl for " + obj.getClass().getName();
				}
			}
			info(message + " : " + toString, throwable);
		}
	}

	public void safeWarn(String message, Object obj)
	{
		if (Level.WARN.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			 warn(message + " : " + toString);
		}
	}

	public void safeWarn(String message, Object obj, Throwable throwable)
	{
		if (Level.WARN.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			warn(message + " : " + toString, throwable);
		}
	}

	public void safeError(String message, Object obj)
	{
		if (Level.ERROR.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			error(message + " : " + toString);
		}
	}

	public void safeError(String message, Object obj, Throwable throwable)
	{
		if (Level.ERROR.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			error(message + " : " + toString, throwable);
		}
	}

	public void safeFatal(String message, Object obj)
	{
		if (Level.FATAL.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			fatal(message + " : " + toString);
		}
	}

	public void safeFatal(String message, Object obj, Throwable throwable)
	{
		if (Level.FATAL.isGreaterOrEqual(getLevel()))
		{
			String toString = null;
			try
			{
				toString = obj.toString();
			}
			catch (Throwable t)
			{
				toString = "BAD toString() impl for " + obj.getClass().getName();
			}
			fatal(message + " : " + toString, throwable);
		}
	}

	public boolean isDebugEnabled()
	{
		return getLogger().isDebugEnabled();
	}

	public boolean isEnabledFor(int priority)
	{
		return getLogger().isEnabledFor((Level) priorityMap.get(new Integer(priority)));
	}

	/**
	 * @see org.apache.ojb.broker.util.configuration.Configurable#configure(Configuration)
	 */
	public void configure(Configuration config) throws ConfigurationException
	{
        if (!isLog4JConfigured())
        {
            LoggingConfiguration lc = (LoggingConfiguration) config;
            initializeLog4JSubSystem(lc.getLoggerConfigFile());
        }
	}

}
